# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
    Implements ObjectID base class and global object registry.

    It used to be that we could store the HDF5 identifier in an ObjectID
    and simply close it when the object was deallocated.  However, since
    HDF5 1.8.5 they have started recycling object identifiers, which
    breaks this system.

    We now use a global registry of object identifiers.  This is implemented
    via a dictionary which maps an integer representation of the identifier
    to a weak reference of an ObjectID.  There is only one ObjectID instance
    in the universe for each integer identifier.  When the HDF5 reference
    count for a identifier reaches zero, HDF5 closes the object and reclaims
    the identifier. When this occurs, the identifier and weak reference must
    be deleted from the registry. If an ObjectID is deallocated, it is deleted
    from the registry and the HDF5 reference count is decreased, HDF5 closes
    and reclaims the identifier for future use.

    All interactions with the registry must be synchronized for thread safety.
    You must acquire "registry.lock" before interacting with the registry. The
    registry is not internally synchronized, in the interest of performance: we
    don't want the same thread attempting to acquire the lock multiple times
    during a single operation, if we can avoid it.

    All ObjectIDs and subclasses thereof should be opened with the "open"
    classmethod factory function, such that an existing ObjectID instance can
    be returned from the registry when appropriate.
"""

from defs cimport *

from weakref import KeyedRef, ref

## {{{ http://code.activestate.com/recipes/577336/ (r3)
from cpython cimport pythread
from cpython.exc cimport PyErr_NoMemory

cdef class FastRLock:
    """Fast, re-entrant locking.

    Under uncongested conditions, the lock is never acquired but only
    counted.  Only when a second thread comes in and notices that the
    lock is needed, it acquires the lock and notifies the first thread
    to release it when it's done.  This is all made possible by the
    wonderful GIL.
    """
    cdef pythread.PyThread_type_lock _real_lock
    cdef long _owner            # ID of thread owning the lock
    cdef int _count             # re-entry count
    cdef int _pending_requests  # number of pending requests for real lock
    cdef bint _is_locked        # whether the real lock is acquired

    def __cinit__(self):
        self._owner = -1
        self._count = 0
        self._is_locked = False
        self._pending_requests = 0
        self._real_lock = pythread.PyThread_allocate_lock()
        if self._real_lock is NULL:
            PyErr_NoMemory()

    def __dealloc__(self):
        if self._real_lock is not NULL:
            pythread.PyThread_free_lock(self._real_lock)
            self._real_lock = NULL

    def acquire(self, bint blocking=True):
        return lock_lock(self, pythread.PyThread_get_thread_ident(), blocking)

    def release(self):
        if self._owner != pythread.PyThread_get_thread_ident():
            raise RuntimeError("cannot release un-acquired lock")
        unlock_lock(self)

    # compatibility with threading.RLock

    def __enter__(self):
        # self.acquire()
        return lock_lock(self, pythread.PyThread_get_thread_ident(), True)

    def __exit__(self, t, v, tb):
        # self.release()
        if self._owner != pythread.PyThread_get_thread_ident():
            raise RuntimeError("cannot release un-acquired lock")
        unlock_lock(self)

    def _is_owned(self):
        return self._owner == pythread.PyThread_get_thread_ident()


cdef inline bint lock_lock(FastRLock lock, long current_thread, bint blocking) nogil:
    # Note that this function *must* hold the GIL when being called.
    # We just use 'nogil' in the signature to make sure that no Python
    # code execution slips in that might free the GIL

    if lock._count:
        # locked! - by myself?
        if current_thread == lock._owner:
            lock._count += 1
            return 1
    elif not lock._pending_requests:
        # not locked, not requested - go!
        lock._owner = current_thread
        lock._count = 1
        return 1
    # need to get the real lock
    return _acquire_lock(
        lock, current_thread,
        pythread.WAIT_LOCK if blocking else pythread.NOWAIT_LOCK)

cdef bint _acquire_lock(FastRLock lock, long current_thread, int wait) nogil:
    # Note that this function *must* hold the GIL when being called.
    # We just use 'nogil' in the signature to make sure that no Python
    # code execution slips in that might free the GIL

    if not lock._is_locked and not lock._pending_requests:
        # someone owns it but didn't acquire the real lock - do that
        # now and tell the owner to release it when done. Note that we
        # do not release the GIL here as we must absolutely be the one
        # who acquires the lock now.
        if not pythread.PyThread_acquire_lock(lock._real_lock, wait):
            return 0
        #assert not lock._is_locked
        lock._is_locked = True
    lock._pending_requests += 1
    with nogil:
        # wait for the lock owning thread to release it
        locked = pythread.PyThread_acquire_lock(lock._real_lock, wait)
    lock._pending_requests -= 1
    #assert not lock._is_locked
    #assert lock._count == 0
    if not locked:
        return 0
    lock._is_locked = True
    lock._owner = current_thread
    lock._count = 1
    return 1

cdef inline void unlock_lock(FastRLock lock) nogil:
    # Note that this function *must* hold the GIL when being called.
    # We just use 'nogil' in the signature to make sure that no Python
    # code execution slips in that might free the GIL

    #assert lock._owner == pythread.PyThread_get_thread_ident()
    #assert lock._count > 0
    lock._count -= 1
    if lock._count == 0:
        lock._owner = -1
        if lock._is_locked:
            pythread.PyThread_release_lock(lock._real_lock)
            lock._is_locked = False
## end of http://code.activestate.com/recipes/577336/ }}}


cdef class _Registry:

    cdef object _data
    cdef readonly FastRLock lock

    def __cinit__(self):
        self._data = {}
        self.lock = FastRLock()

    __hash__ = None # Avoid Py3 warning

    def cleanup(self):
        "Manage invalid identifiers"
        deadlist = []
        for key in self._data:
            val = self._data[key]
            val = val()
            if val is None:
                deadlist.append(key)
                continue
            if not val.valid:
                deadlist.append(key)
        for key in deadlist:
            del self._data[key]

    def __getitem__(self, key):
        o = self._data[key]()
        if o is None:
            # This would occur if we had open objects and closed their
            # file, causing the objects identifiers to be reclaimed.
            # Now we clean up the registry when we close a file (or any
            # other identifier, for that matter), so in practice this
            # condition never obtains.
            del self._data[key]
            # We need to raise a KeyError:
            o = self._data[key]()
        return o

    def __setitem__(self, key, val):
        # this method should only be called by ObjectID.open
        self._data[key] = ref(val)

    def __delitem__(self, key):
        # we need to synchronize removal of the id from the
        # registry with decreasing the HDF5 reference count:
        self._data.pop(key,None)

        if H5Iget_type(key) >= 0: # if not, object was explicitly closed
            H5Idec_ref(key) 


registry = _Registry()


cdef class ObjectID:

    """
        Represents an HDF5 identifier.

    """

    property fileno:
        def __get__(self):
            cdef H5G_stat_t stat
            H5Gget_objinfo(self.id, '.', 0, &stat)
            return (stat.fileno[0], stat.fileno[1])

    property valid:
        def __get__(self):
            if not self.id:
                return False
            res = H5Iget_type(self.id) > 0
            if not res:
                self.id = 0
            return res

    def __cinit__(self, id):
        self.id = id
        self.locked = 0
        with registry.lock:
            registry[id] = self

    def __dealloc__(self):
        if not self.locked:
            try:
                with registry.lock:
                    del registry[self.id]
            except AttributeError:
                # library being torn down, registry is None
                pass

    def __nonzero__(self):
        return self.valid

    def __copy__(self):
        cdef ObjectID cpy
        cpy = type(self)(self.id)
        return cpy

    def __richcmp__(self, object other, int how):
        """ Default comparison mechanism for HDF5 objects (equal/not-equal)

        Default equality testing:
        1. Objects which are not both ObjectIDs are unequal
        2. Objects with the same HDF5 ID number are always equal
        3. Objects which hash the same are equal
        """
        cdef bint equal = 0

        if how != 2 and how != 3:
            return NotImplemented

        if isinstance(other, ObjectID):
            if self.id == other.id:
                equal = 1
            else:
                try:
                    equal = hash(self) == hash(other)
                except TypeError:
                    pass

        if how == 2:
            return equal
        return not equal

    def __hash__(self):
        """ Default hashing mechanism for HDF5 objects

        Default hashing strategy:
        1. Try to hash based on the object's fileno and objno records
        2. If (1) succeeds, cache the resulting value
        3. If (1) fails, raise TypeError
        """
        cdef H5G_stat_t stat

        if self._hash is None:
            try:
                H5Gget_objinfo(self.id, '.', 0, &stat)
                self._hash = hash((stat.fileno[0], stat.fileno[1], stat.objno[0], stat.objno[1]))
            except Exception:
                raise TypeError("Objects of class %s cannot be hashed" % self.__class__.__name__)

        return self._hash

    @classmethod
    def open(cls, id):
        """ Return a representation of an HDF5 identifier """
        with registry.lock:
            try:
                res = registry[id]
            except KeyError:
                res = cls(id)
            return res


cdef hid_t pdefault(ObjectID pid):

    if pid is None:
        return <hid_t>H5P_DEFAULT
    return pid.id



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

# This file is based on code from the PyTables project.  The complete PyTables
# license is available at licenses/pytables.txt, in the distribution root
# directory.

# API for NumPy objects
cdef extern from "numpy/arrayobject.h":

  # Platform independent types
  ctypedef int npy_intp
  ctypedef signed int npy_int8
  ctypedef unsigned int npy_uint8
  ctypedef signed int npy_int16
  ctypedef unsigned int npy_uint16
  ctypedef signed int npy_int32
  ctypedef unsigned int npy_uint32
  ctypedef signed long long npy_int64
  ctypedef unsigned long long npy_uint64
  ctypedef float npy_float32
  ctypedef double npy_float64

  cdef enum NPY_TYPES:
    NPY_BOOL
    NPY_BYTE
    NPY_UBYTE
    NPY_SHORT
    NPY_USHORT
    NPY_INT
    NPY_UINT
    NPY_LONG
    NPY_ULONG
    NPY_LONGLONG
    NPY_ULONGLONG
    NPY_FLOAT
    NPY_DOUBLE
    NPY_LONGDOUBLE
    NPY_CFLOAT
    NPY_CDOUBLE
    NPY_CLONGDOUBLE
    NPY_OBJECT
    NPY_STRING
    NPY_UNICODE
    NPY_VOID
    NPY_NTYPES
    NPY_NOTYPE

  # Platform independent types
  cdef enum:
    NPY_INT8, NPY_INT16, NPY_INT32, NPY_INT64,
    NPY_UINT8, NPY_UINT16, NPY_UINT32, NPY_UINT64,
    NPY_FLOAT32, NPY_FLOAT64, NPY_COMPLEX64, NPY_COMPLEX128

  cdef enum:
    NPY_WRITEABLE, NPY_ALIGNED, NPY_C_CONTIGUOUS, NPY_CONTIGUOUS,
    NPY_FORCECAST, NPY_NOTSWAPPED, NPY_OWNDATA

  # Classes
  ctypedef extern class numpy.dtype [object PyArray_Descr]:
    cdef int type_num, elsize, alignment
    cdef char type, kind, byteorder, hasobject
    cdef object fields, typeobj

  ctypedef extern class numpy.ndarray [object PyArrayObject]:
    cdef char *data
    cdef int nd
    cdef npy_intp *dimensions
    cdef npy_intp *strides
    cdef object base
    cdef dtype descr
    cdef int flags

  ctypedef struct npy_cfloat:
    float real
    float imag

  ctypedef struct npy_cdouble:
    double real
    double imag

  # Functions
  int PyArray_DIM(ndarray arr, int i)
  object PyArray_FROM_OF(object arr, int requirements)

  object PyArray_GETITEM(object arr, void *itemptr)
  int PyArray_SETITEM(object arr, void *itemptr, object obj)
  dtype PyArray_DescrFromType(int type)
  object PyArray_Scalar(void *data, dtype descr, object base)
  long PyArray_NBYTES(object arr)

  int PyArray_CheckScalar(object sclr)
  void PyArray_ScalarAsCtype(object sclr, void* ptr)
  object PyArray_SimpleNew(int nd, npy_intp* dims, int typenum)
  object PyArray_ContiguousFromAny(object arr, int typenum, int min_depth, int max_depth)
  object PyArray_FROM_OTF(object arr, int typenum, int requirements)

  # The NumPy initialization function
  void import_array()

  void* PyArray_DATA(ndarray arr)








# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.
"""
    Provides access to the low-level HDF5 "H5D" dataset interface.
"""

# Compile-time imports
from _objects cimport pdefault
from numpy cimport ndarray, import_array, PyArray_DATA, NPY_WRITEABLE
from utils cimport  check_numpy_read, check_numpy_write, \
                    convert_tuple, emalloc, efree
from h5t cimport TypeID, typewrap, py_create
from h5s cimport SpaceID
from h5p cimport PropID, propwrap
from _proxy cimport dset_rw

from h5py import _objects

# Initialization
import_array()

# === Public constants and data structures ====================================

COMPACT     = H5D_COMPACT
CONTIGUOUS  = H5D_CONTIGUOUS
CHUNKED     = H5D_CHUNKED

ALLOC_TIME_DEFAULT  = H5D_ALLOC_TIME_DEFAULT
ALLOC_TIME_LATE     = H5D_ALLOC_TIME_LATE
ALLOC_TIME_EARLY    = H5D_ALLOC_TIME_EARLY
ALLOC_TIME_INCR     = H5D_ALLOC_TIME_INCR

SPACE_STATUS_NOT_ALLOCATED  = H5D_SPACE_STATUS_NOT_ALLOCATED
SPACE_STATUS_PART_ALLOCATED = H5D_SPACE_STATUS_PART_ALLOCATED
SPACE_STATUS_ALLOCATED      = H5D_SPACE_STATUS_ALLOCATED

FILL_TIME_ALLOC = H5D_FILL_TIME_ALLOC
FILL_TIME_NEVER = H5D_FILL_TIME_NEVER
FILL_TIME_IFSET = H5D_FILL_TIME_IFSET

FILL_VALUE_UNDEFINED    = H5D_FILL_VALUE_UNDEFINED
FILL_VALUE_DEFAULT      = H5D_FILL_VALUE_DEFAULT
FILL_VALUE_USER_DEFINED = H5D_FILL_VALUE_USER_DEFINED


# === Dataset operations ======================================================

def create(ObjectID loc not None, object name, TypeID tid not None,
               SpaceID space not None, PropID dcpl=None, PropID lcpl=None, PropID dapl = None):
        """ (objectID loc, STRING name or None, TypeID tid, SpaceID space,
             PropDCID dcpl=None, PropID lcpl=None) => DatasetID

        Create a new dataset.  If "name" is None, the dataset will be
        anonymous.
        """
        cdef hid_t dsid
        cdef char* cname = NULL
        if name is not None:
            cname = name

        if cname != NULL:
            dsid = H5Dcreate2(loc.id, cname, tid.id, space.id,
                     pdefault(lcpl), pdefault(dcpl), pdefault(dapl))
        else:
            dsid = H5Dcreate_anon(loc.id, tid.id, space.id,
                     pdefault(dcpl), pdefault(dapl))
        return DatasetID.open(dsid)

def open(ObjectID loc not None, char* name):
    """ (ObjectID loc, STRING name) => DatasetID

        Open an existing dataset attached to a group or file object, by name.
    """
    return DatasetID.open(H5Dopen(loc.id, name))

# --- Proxy functions for safe(r) threading -----------------------------------


cdef class DatasetID(ObjectID):

    """
        Represents an HDF5 dataset identifier.

        Objects of this class may be used in any HDF5 function which expects
        a dataset identifier.  Also, all H5D* functions which take a dataset
        instance as their first argument are presented as methods of this
        class.

        Properties:
        dtype:  Numpy dtype representing the dataset type
        shape:  Numpy-style shape tuple representing the dataspace
        rank:   Integer giving dataset rank

        * Hashable: Yes, unless anonymous
        * Equality: True HDF5 identity if unless anonymous
    """

    property dtype:
        """ Numpy dtype object representing the dataset type """
        def __get__(self):
            # Dataset type can't change
            cdef TypeID tid
            if self._dtype is None:
                tid = self.get_type()
                self._dtype = tid.dtype
            return self._dtype

    property shape:
        """ Numpy-style shape tuple representing the dataspace """
        def __get__(self):
            # Shape can change (DatasetID.extend), so don't cache it
            cdef SpaceID sid
            sid = self.get_space()
            return sid.get_simple_extent_dims()

    property rank:
        """ Integer giving the dataset rank (0 = scalar) """
        def __get__(self):
            cdef SpaceID sid
            sid = self.get_space()
            return sid.get_simple_extent_ndims()


    def _close(self):
        """ ()

            Terminate access through this identifier.  You shouldn't have to
            call this manually; Dataset objects are automatically destroyed
            when their Python wrappers are freed.
        """
        with _objects.registry.lock:
            H5Dclose(self.id)
            if not self.valid:
                del _objects.registry[self.id]


    def read(self, SpaceID mspace not None, SpaceID fspace not None,
                   ndarray arr_obj not None, TypeID mtype=None,
                   PropID dxpl=None):
        """ (SpaceID mspace, SpaceID fspace, NDARRAY arr_obj,
             TypeID mtype=None, PropDXID dxpl=None)

            Read data from an HDF5 dataset into a Numpy array.

            It is your responsibility to ensure that the memory dataspace
            provided is compatible with the shape of the Numpy array.  Since a
            wide variety of dataspace configurations are possible, this is not
            checked.  You can easily crash Python by reading in data from too
            large a dataspace.

            If a memory datatype is not specified, one will be auto-created
            based on the array's dtype.

            The provided Numpy array must be writable and C-contiguous.  If
            this is not the case, ValueError will be raised and the read will
            fail.  Keyword dxpl may be a dataset transfer property list.
        """
        cdef hid_t self_id, mtype_id, mspace_id, fspace_id, plist_id
        cdef void* data
        cdef int oldflags

        if mtype is None:
            mtype = py_create(arr_obj.dtype)
        check_numpy_write(arr_obj, -1)

        self_id = self.id
        mtype_id = mtype.id
        mspace_id = mspace.id
        fspace_id = fspace.id
        plist_id = pdefault(dxpl)
        data = PyArray_DATA(arr_obj)

        dset_rw(self_id, mtype_id, mspace_id, fspace_id, plist_id, data, 1)

    def write(self, SpaceID mspace not None, SpaceID fspace not None,
                    ndarray arr_obj not None, TypeID mtype=None,
                    PropID dxpl=None):
        """ (SpaceID mspace, SpaceID fspace, NDARRAY arr_obj,
             TypeID mtype=None, PropDXID dxpl=None)

            Write data from a Numpy array to an HDF5 dataset. Keyword dxpl may
            be a dataset transfer property list.

            It is your responsibility to ensure that the memory dataspace
            provided is compatible with the shape of the Numpy array.  Since a
            wide variety of dataspace configurations are possible, this is not
            checked.  You can easily crash Python by writing data from too
            large a dataspace.

            If a memory datatype is not specified, one will be auto-created
            based on the array's dtype.

            The provided Numpy array must be C-contiguous.  If this is not the
            case, ValueError will be raised and the read will fail.
        """
        cdef hid_t self_id, mtype_id, mspace_id, fspace_id, plist_id
        cdef void* data
        cdef int oldflags

        if mtype is None:
            mtype = py_create(arr_obj.dtype)
        check_numpy_read(arr_obj, -1)

        self_id = self.id
        mtype_id = mtype.id
        mspace_id = mspace.id
        fspace_id = fspace.id
        plist_id = pdefault(dxpl)
        data = PyArray_DATA(arr_obj)

        dset_rw(self_id, mtype_id, mspace_id, fspace_id, plist_id, data, 0)

    def extend(self, tuple shape):
        """ (TUPLE shape)

            Extend the given dataset so it's at least as big as "shape".  Note
            that a dataset may only be extended up to the maximum dimensions of
            its dataspace, which are fixed when the dataset is created.
        """
        cdef int rank
        cdef hid_t space_id = 0
        cdef hsize_t* dims = NULL

        try:
            space_id = H5Dget_space(self.id)
            rank = H5Sget_simple_extent_ndims(space_id)

            if len(shape) != rank:
                raise TypeError("New shape length (%d) must match dataset rank (%d)" % (len(shape), rank))

            dims = <hsize_t*>emalloc(sizeof(hsize_t)*rank)
            convert_tuple(shape, dims, rank)
            H5Dextend(self.id, dims)

        finally:
            efree(dims)
            if space_id:
                H5Sclose(space_id)


    def set_extent(self, tuple shape):
        """ (TUPLE shape)

            Set the size of the dataspace to match the given shape.  If the new
            size is larger in any dimension, it must be compatible with the
            maximum dataspace size.
        """
        cdef int rank
        cdef hid_t space_id = 0
        cdef hsize_t* dims = NULL

        try:
            space_id = H5Dget_space(self.id)
            rank = H5Sget_simple_extent_ndims(space_id)

            if len(shape) != rank:
                raise TypeError("New shape length (%d) must match dataset rank (%d)" % (len(shape), rank))

            dims = <hsize_t*>emalloc(sizeof(hsize_t)*rank)
            convert_tuple(shape, dims, rank)
            H5Dset_extent(self.id, dims)

        finally:
            efree(dims)
            if space_id:
                H5Sclose(space_id)



    def get_space(self):
        """ () => SpaceID

            Create and return a new copy of the dataspace for this dataset.
        """
        return SpaceID.open(H5Dget_space(self.id))


    def get_space_status(self):
        """ () => INT space_status_code

            Determine if space has been allocated for a dataset.
            Return value is one of:

            * SPACE_STATUS_NOT_ALLOCATED
            * SPACE_STATUS_PART_ALLOCATED
            * SPACE_STATUS_ALLOCATED
        """
        cdef H5D_space_status_t status
        H5Dget_space_status(self.id, &status)
        return <int>status


    def get_type(self):
        """ () => TypeID

            Create and return a new copy of the datatype for this dataset.
        """
        return typewrap(H5Dget_type(self.id))


    def get_create_plist(self):
        """ () => PropDCID

            Create an return a new copy of the dataset creation property list
            used when this dataset was created.
        """
        return propwrap(H5Dget_create_plist(self.id))


    def get_offset(self):
        """ () => LONG offset or None

            Get the offset of this dataset in the file, in bytes, or None if
            it doesn't have one.  This is always the case for datasets which
            use chunked storage, compact datasets, and datasets for which space
            has not yet been allocated in the file.
        """
        cdef haddr_t offset
        offset = H5Dget_offset(self.id)
        if offset == HADDR_UNDEF:
            return None
        return offset


    def get_storage_size(self):
        """ () => LONG storage_size

            Determine the amount of file space required for a dataset.  Note
            this only counts the space which has actually been allocated; it
            may even be zero.
        """
        return H5Dget_storage_size(self.id)



# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

"""
    Provides access to the low-level HDF5 "H5A" attribute interface.
"""

# Compile-time imports
from _objects cimport pdefault
from h5t cimport TypeID, typewrap, py_create
from h5s cimport SpaceID
from h5p cimport PropID
from numpy cimport import_array, ndarray, PyArray_DATA
from utils cimport check_numpy_read, check_numpy_write, emalloc, efree
from _proxy cimport attr_rw

from h5py import _objects

# Initialization
import_array()

# === General attribute operations ============================================

# --- create, create_by_name ---

def create(ObjectID loc not None, char* name, TypeID tid not None,
    SpaceID space not None, *, char* obj_name='.', PropID lapl=None):
    """(ObjectID loc, STRING name, TypeID tid, SpaceID space, **kwds) => AttrID

    Create a new attribute, attached to an existing object.

    STRING obj_name (".")
        Attach attribute to this group member instead

    PropID lapl
        Link access property list for obj_name
    """

    return AttrID.open(H5Acreate_by_name(loc.id, obj_name, name, tid.id,
            space.id, H5P_DEFAULT, H5P_DEFAULT, pdefault(lapl)))


# --- open, open_by_name, open_by_idx ---

def open(ObjectID loc not None, char* name=NULL, int index=-1, *,
    char* obj_name='.', int index_type=H5_INDEX_NAME, int order=H5_ITER_NATIVE,
    PropID lapl=None):
    """(ObjectID loc, STRING name=, INT index=, **kwds) => AttrID

    Open an attribute attached to an existing object.  You must specify
    exactly one of either name or idx.  Keywords are:

    STRING obj_name (".")
        Attribute is attached to this group member

    PropID lapl (None)
        Link access property list for obj_name

    INT index_type (h5.INDEX_NAME)

    INT order (h5.ITER_NATIVE)

    """
    if (name == NULL and index < 0) or (name != NULL and index >= 0):
        raise TypeError("Exactly one of name or idx must be specified")

    if name != NULL:
        return AttrID.open(H5Aopen_by_name(loc.id, obj_name, name,
                        H5P_DEFAULT, pdefault(lapl)))
    else:
        return AttrID.open(H5Aopen_by_idx(loc.id, obj_name,
            <H5_index_t>index_type, <H5_iter_order_t>order, index,
            H5P_DEFAULT, pdefault(lapl)))


# --- exists, exists_by_name ---

def exists(ObjectID loc not None, char* name, *,
            char* obj_name=".", PropID lapl=None):
    """(ObjectID loc, STRING name, **kwds) => BOOL

    Determine if an attribute is attached to this object.  Keywords:

    STRING obj_name (".")
        Look for attributes attached to this group member

    PropID lapl (None):
        Link access property list for obj_name
    """
    return <bint>H5Aexists_by_name(loc.id, obj_name, name, pdefault(lapl))


# --- rename, rename_by_name ---

def rename(ObjectID loc not None, char* name, char* new_name, *,
    char* obj_name='.', PropID lapl=None):
    """(ObjectID loc, STRING name, STRING new_name, **kwds)

    Rename an attribute.  Keywords:

    STRING obj_name (".")
        Attribute is attached to this group member

    PropID lapl (None)
        Link access property list for obj_name
    """
    H5Arename_by_name(loc.id, obj_name, name, new_name, pdefault(lapl))


def delete(ObjectID loc not None, char* name=NULL, int index=-1, *,
    char* obj_name='.', int index_type=H5_INDEX_NAME, int order=H5_ITER_NATIVE,
    PropID lapl=None):
    """(ObjectID loc, STRING name=, INT index=, **kwds)

    Remove an attribute from an object.  Specify exactly one of "name"
    or "index". Keyword-only arguments:

    STRING obj_name (".")
        Attribute is attached to this group member

    PropID lapl (None)
        Link access property list for obj_name

    INT index_type (h5.INDEX_NAME)

    INT order (h5.ITER_NATIVE)
    """
    if name != NULL and index < 0:
        H5Adelete_by_name(loc.id, obj_name, name, pdefault(lapl))
    elif name == NULL and index >= 0:
        H5Adelete_by_idx(loc.id, obj_name, <H5_index_t>index_type,
            <H5_iter_order_t>order, index, pdefault(lapl))
    else:
        raise TypeError("Exactly one of index or name must be specified.")

def get_num_attrs(ObjectID loc not None):
    """(ObjectID loc) => INT

    Determine the number of attributes attached to an HDF5 object.
    """
    return H5Aget_num_attrs(loc.id)


cdef class AttrInfo:

    cdef H5A_info_t info

    property corder_valid:
        """Indicates if the creation order is valid"""
        def __get__(self):
            return <bint>self.info.corder_valid
    property corder:
        """Creation order"""
        def __get__(self):
            return <int>self.info.corder
    property cset:
        """Character set of attribute name (integer typecode from h5t)"""
        def __get__(self):
            return <int>self.info.cset
    property data_size:
        """Size of raw data"""
        def __get__(self):
            return self.info.data_size

    def _hash(self):
        return hash((self.corder_valid, self.corder, self.cset, self.data_size))


def get_info(ObjectID loc not None, char* name=NULL, int index=-1, *,
            char* obj_name='.', PropID lapl=None,
            int index_type=H5_INDEX_NAME, int order=H5_ITER_NATIVE):
    """(ObjectID loc, STRING name=, INT index=, **kwds) => AttrInfo

    Get information about an attribute, in one of two ways:

    1. If you have the attribute identifier, just pass it in
    2. If you have the parent object, supply it and exactly one of
       either name or index.

    STRING obj_name (".")
        Use this group member instead

    PropID lapl (None)
        Link access property list for obj_name

    INT index_type (h5.INDEX_NAME)
        Which index to use

    INT order (h5.ITER_NATIVE)
        What order the index is in
    """
    cdef AttrInfo info = AttrInfo()

    if name == NULL and index < 0:
        H5Aget_info(loc.id, &info.info)
    elif name != NULL and index >= 0:
        raise TypeError("At most one of name and index may be specified")
    elif name != NULL:
        H5Aget_info_by_name(loc.id, obj_name, name, &info.info, pdefault(lapl))
    elif index >= 0:
        H5Aget_info_by_idx(loc.id, obj_name, <H5_index_t>index_type,
            <H5_iter_order_t>order, index, &info.info, pdefault(lapl))

    return info

# === Iteration routines ======================================================

cdef class _AttrVisitor:
    cdef object func
    cdef object retval
    def __init__(self, func):
        self.func = func
        self.retval = None

cdef herr_t cb_attr_iter(hid_t loc_id, char* attr_name, H5A_info_t *ainfo, void* vis_in) except 2:
    cdef _AttrVisitor vis = <_AttrVisitor>vis_in
    cdef AttrInfo info = AttrInfo()
    info.info = ainfo[0]
    vis.retval = vis.func(attr_name, info)
    if vis.retval is not None:
        return 1
    return 0

cdef herr_t cb_attr_simple(hid_t loc_id, char* attr_name, H5A_info_t *ainfo, void* vis_in) except 2:
    cdef _AttrVisitor vis = <_AttrVisitor>vis_in
    vis.retval = vis.func(attr_name)
    if vis.retval is not None:
        return 1
    return 0


def iterate(ObjectID loc not None, object func, int index=0, *,
    int index_type=H5_INDEX_NAME, int order=H5_ITER_NATIVE, bint info=0):
    """(ObjectID loc, CALLABLE func, INT index=0, **kwds) => <Return value from func>

    Iterate a callable (function, method or callable object) over the
    attributes attached to this object.  You callable should have the
    signature::

        func(STRING name) => Result

    or if the keyword argument "info" is True::

        func(STRING name, AttrInfo info) => Result

    Returning None continues iteration; returning anything else aborts
    iteration and returns that value.  Keywords:

    BOOL info (False)
        Callback is func(STRING name, AttrInfo info), not func(STRING name)

    INT index_type (h5.INDEX_NAME)
        Which index to use

    INT order (h5.ITER_NATIVE)
        Index order to use
    """
    if index < 0:
        raise ValueError("Starting index must be a non-negative integer.")

    cdef hsize_t i = index
    cdef _AttrVisitor vis = _AttrVisitor(func)
    cdef H5A_operator2_t cfunc

    if info:
        cfunc = cb_attr_iter
    else:
        cfunc = cb_attr_simple

    H5Aiterate2(loc.id, <H5_index_t>index_type, <H5_iter_order_t>order,
        &i, cfunc, <void*>vis)

    return vis.retval



# === Attribute class & methods ===============================================

cdef class AttrID(ObjectID):

    """
        Logical representation of an HDF5 attribute identifier.

        Objects of this class can be used in any HDF5 function call
        which expects an attribute identifier.  Additionally, all ``H5A*``
        functions which always take an attribute instance as the first
        argument are presented as methods of this class.

        * Hashable: No
        * Equality: Identifier comparison
    """

    property name:
        """The attribute's name"""
        def __get__(self):
            return self.get_name()

    property shape:
        """A Numpy-style shape tuple representing the attribute's dataspace"""
        def __get__(self):

            cdef SpaceID space
            space = self.get_space()
            return space.get_simple_extent_dims()

    property dtype:
        """A Numpy-stype dtype object representing the attribute's datatype"""
        def __get__(self):

            cdef TypeID tid
            tid = self.get_type()
            return tid.py_dtype()


    def _close(self):
        """()

        Close this attribute and release resources.  You don't need to
        call this manually; attributes are automatically destroyed when
        their Python wrappers are freed.
        """
        with _objects.registry.lock:
            H5Aclose(self.id)
            if not self.valid:
                del _objects.registry[self.id]


    def read(self, ndarray arr not None):
        """(NDARRAY arr)

        Read the attribute data into the given Numpy array.  Note that the
        Numpy array must have the same shape as the HDF5 attribute, and a
        conversion-compatible datatype.

        The Numpy array must be writable and C-contiguous.  If this is not
        the case, the read will fail with an exception.
        """
        cdef TypeID mtype
        cdef hid_t space_id
        space_id = 0

        try:
            space_id = H5Aget_space(self.id)
            check_numpy_write(arr, space_id)

            mtype = py_create(arr.dtype)

            attr_rw(self.id, mtype.id, PyArray_DATA(arr), 1)

        finally:
            if space_id:
                H5Sclose(space_id)


    def write(self, ndarray arr not None):
        """(NDARRAY arr)

        Write the contents of a Numpy array too the attribute.  Note that
        the Numpy array must have the same shape as the HDF5 attribute, and
        a conversion-compatible datatype.

        The Numpy array must be C-contiguous.  If this is not the case,
        the write will fail with an exception.
        """
        cdef TypeID mtype
        cdef hid_t space_id
        space_id = 0

        try:
            space_id = H5Aget_space(self.id)
            check_numpy_read(arr, space_id)
            mtype = py_create(arr.dtype)

            attr_rw(self.id, mtype.id, PyArray_DATA(arr), 0)

        finally:
            if space_id:
                H5Sclose(space_id)


    def get_name(self):
        """() => STRING name

        Determine the name of this attribute.
        """
        cdef int blen
        cdef char* buf
        buf = NULL

        try:
            blen = H5Aget_name(self.id, 0, NULL)
            assert blen >= 0
            buf = <char*>emalloc(sizeof(char)*blen+1)
            blen = H5Aget_name(self.id, blen+1, buf)
            strout = <bytes>buf
        finally:
            efree(buf)

        return strout


    def get_space(self):
        """() => SpaceID

        Create and return a copy of the attribute's dataspace.
        """
        return SpaceID.open(H5Aget_space(self.id))


    def get_type(self):
        """() => TypeID

        Create and return a copy of the attribute's datatype.
        """
        return typewrap(H5Aget_type(self.id))


    def get_storage_size(self):
        """() => INT

        Get the amount of storage required for this attribute.
        """
        return H5Aget_storage_size(self.id)











include "config.pxi"
from api_types_ext cimport *
from api_types_hdf5 cimport *

cimport _hdf5

from _errors cimport set_exception

include "_locks.pxi"

rlock = FastRLock()
cdef herr_t H5open() except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5open()
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5close() except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5close()
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5get_libversion(unsigned *majnum, unsigned *minnum, unsigned *relnum) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5get_libversion(majnum, minnum, relnum)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Dcreate2(hid_t loc_id, char *name, hid_t type_id, hid_t space_id, hid_t lcpl_id, hid_t dcpl_id, hid_t dapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Dcreate2(loc_id, name, type_id, space_id, lcpl_id, dcpl_id, dapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Dcreate_anon(hid_t file_id, hid_t type_id, hid_t space_id, hid_t plist_id, hid_t dapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Dcreate_anon(file_id, type_id, space_id, plist_id, dapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Dopen(hid_t file_id, char *name) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Dopen(file_id, name)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Dclose(hid_t dset_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dclose(dset_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Dget_space(hid_t dset_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Dget_space(dset_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Dget_space_status(hid_t dset_id, H5D_space_status_t *status) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dget_space_status(dset_id, status)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Dget_type(hid_t dset_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Dget_type(dset_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Dget_create_plist(hid_t dataset_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Dget_create_plist(dataset_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef haddr_t H5Dget_offset(hid_t dset_id) except *:
    cdef haddr_t r
    with rlock:
        r = _hdf5.H5Dget_offset(dset_id)
        if r==0:
            if set_exception():
                return <haddr_t>0;
        return r

cdef hsize_t H5Dget_storage_size(hid_t dset_id) except *:
    cdef hsize_t r
    with rlock:
        r = _hdf5.H5Dget_storage_size(dset_id)
        if r==0:
            if set_exception():
                return <hsize_t>0;
        return r

cdef herr_t H5Dread(hid_t dset_id, hid_t mem_type_id, hid_t mem_space_id, hid_t file_space_id, hid_t plist_id, void *buf) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dread(dset_id, mem_type_id, mem_space_id, file_space_id, plist_id, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Dwrite(hid_t dset_id, hid_t mem_type, hid_t mem_space, hid_t file_space, hid_t xfer_plist, void* buf) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dwrite(dset_id, mem_type, mem_space, file_space, xfer_plist, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Dextend(hid_t dataset_id, hsize_t *size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dextend(dataset_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Dfill(void *fill, hid_t fill_type_id, void *buf,  hid_t buf_type_id, hid_t space_id ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dfill(fill, fill_type_id, buf, buf_type_id, space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Dvlen_get_buf_size(hid_t dset_id, hid_t type_id, hid_t space_id, hsize_t *size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dvlen_get_buf_size(dset_id, type_id, space_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Dvlen_reclaim(hid_t type_id, hid_t space_id,  hid_t plist, void *buf) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dvlen_reclaim(type_id, space_id, plist, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Diterate(void *buf, hid_t type_id, hid_t space_id,  H5D_operator_t op, void* operator_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Diterate(buf, type_id, space_id, op, operator_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Dset_extent(hid_t dset_id, hsize_t* size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Dset_extent(dset_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Fcreate(char *filename, unsigned int flags, hid_t create_plist, hid_t access_plist) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Fcreate(filename, flags, create_plist, access_plist)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Fopen(char *name, unsigned flags, hid_t access_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Fopen(name, flags, access_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Fclose(hid_t file_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fclose(file_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Fis_hdf5(char *name) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Fis_hdf5(name)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Fflush(hid_t object_id, H5F_scope_t scope) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fflush(object_id, scope)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Freopen(hid_t file_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Freopen(file_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Fmount(hid_t loc_id, char *name, hid_t child_id, hid_t plist_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fmount(loc_id, name, child_id, plist_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Funmount(hid_t loc_id, char *name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Funmount(loc_id, name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Fget_filesize(hid_t file_id, hsize_t *size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fget_filesize(file_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Fget_create_plist(hid_t file_id ) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Fget_create_plist(file_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Fget_access_plist(hid_t file_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Fget_access_plist(file_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hssize_t H5Fget_freespace(hid_t file_id) except *:
    cdef hssize_t r
    with rlock:
        r = _hdf5.H5Fget_freespace(file_id)
        if r<0:
            if set_exception():
                return <hssize_t>-1;
        return r

cdef ssize_t H5Fget_name(hid_t obj_id, char *name, size_t size) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Fget_name(obj_id, name, size)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef int H5Fget_obj_count(hid_t file_id, unsigned int types) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Fget_obj_count(file_id, types)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Fget_obj_ids(hid_t file_id, unsigned int types, int max_objs, hid_t *obj_id_list) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Fget_obj_ids(file_id, types, max_objs, obj_id_list)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Fget_vfd_handle(hid_t file_id, hid_t fapl_id, void **file_handle) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fget_vfd_handle(file_id, fapl_id, file_handle)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Fget_intent(hid_t file_id, unsigned int *intent) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fget_intent(file_id, intent)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Fget_mdc_config(hid_t file_id, H5AC_cache_config_t *config_ptr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fget_mdc_config(file_id, config_ptr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Fget_mdc_hit_rate(hid_t file_id, double *hit_rate_ptr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fget_mdc_hit_rate(file_id, hit_rate_ptr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Fget_mdc_size(hid_t file_id, size_t *max_size_ptr, size_t *min_clean_size_ptr, size_t *cur_size_ptr, int *cur_num_entries_ptr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fget_mdc_size(file_id, max_size_ptr, min_clean_size_ptr, cur_size_ptr, cur_num_entries_ptr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Freset_mdc_hit_rate_stats(hid_t file_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Freset_mdc_hit_rate_stats(file_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Fset_mdc_config(hid_t file_id, H5AC_cache_config_t *config_ptr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Fset_mdc_config(file_id, config_ptr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Gcreate(hid_t loc_id, char *name, size_t size_hint) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Gcreate(loc_id, name, size_hint)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Gopen(hid_t loc_id, char *name) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Gopen(loc_id, name)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Gclose(hid_t group_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gclose(group_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Glink2( hid_t curr_loc_id, char *current_name, H5G_link_t link_type, hid_t new_loc_id, char *new_name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Glink2(curr_loc_id, current_name, link_type, new_loc_id, new_name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gunlink(hid_t file_id, char *name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gunlink(file_id, name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gmove2(hid_t src_loc_id, char *src_name, hid_t dst_loc_id, char *dst_name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gmove2(src_loc_id, src_name, dst_loc_id, dst_name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gget_num_objs(hid_t loc_id, hsize_t*  num_obj) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gget_num_objs(loc_id, num_obj)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5Gget_objname_by_idx(hid_t loc_id, hsize_t idx, char *name, size_t size) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Gget_objname_by_idx(loc_id, idx, name, size)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Gget_objtype_by_idx(hid_t loc_id, hsize_t idx) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Gget_objtype_by_idx(loc_id, idx)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Giterate(hid_t loc_id, char *name, int *idx, H5G_iterate_t op, void* data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Giterate(loc_id, name, idx, op, data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gget_objinfo(hid_t loc_id, char* name, int follow_link, H5G_stat_t *statbuf) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gget_objinfo(loc_id, name, follow_link, statbuf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gget_linkval(hid_t loc_id, char *name, size_t size, char *value) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gget_linkval(loc_id, name, size, value)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gset_comment(hid_t loc_id, char *name, char *comment) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gset_comment(loc_id, name, comment)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5Gget_comment(hid_t loc_id, char *name, size_t bufsize, char *comment) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Gget_comment(loc_id, name, bufsize, comment)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef hid_t H5Gcreate_anon( hid_t loc_id, hid_t gcpl_id, hid_t gapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Gcreate_anon(loc_id, gcpl_id, gapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Gcreate2(hid_t loc_id, char *name, hid_t lcpl_id, hid_t gcpl_id, hid_t gapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Gcreate2(loc_id, name, lcpl_id, gcpl_id, gapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Gopen2( hid_t loc_id, char * name, hid_t gapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Gopen2(loc_id, name, gapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Gget_info( hid_t group_id, H5G_info_t *group_info) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gget_info(group_id, group_info)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Gget_info_by_name( hid_t loc_id, char *group_name, H5G_info_t *group_info, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Gget_info_by_name(loc_id, group_name, group_info, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Gget_create_plist(hid_t group_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Gget_create_plist(group_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef H5I_type_t H5Iget_type(hid_t obj_id) except *:
    cdef H5I_type_t r
    with rlock:
        r = _hdf5.H5Iget_type(obj_id)
        if r<0:
            if set_exception():
                return <H5I_type_t>-1;
        return r

cdef ssize_t H5Iget_name( hid_t obj_id, char *name, size_t size) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Iget_name(obj_id, name, size)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef hid_t H5Iget_file_id(hid_t obj_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Iget_file_id(obj_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef int H5Idec_ref(hid_t obj_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Idec_ref(obj_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Iget_ref(hid_t obj_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Iget_ref(obj_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Iinc_ref(hid_t obj_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Iinc_ref(obj_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Lmove(hid_t src_loc, char *src_name, hid_t dst_loc, char *dst_name, hid_t lcpl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lmove(src_loc, src_name, dst_loc, dst_name, lcpl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lcopy(hid_t src_loc, char *src_name, hid_t dst_loc, char *dst_name, hid_t lcpl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lcopy(src_loc, src_name, dst_loc, dst_name, lcpl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lcreate_hard(hid_t cur_loc, char *cur_name, hid_t dst_loc, char *dst_name, hid_t lcpl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lcreate_hard(cur_loc, cur_name, dst_loc, dst_name, lcpl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lcreate_soft(char *link_target, hid_t link_loc_id, char *link_name, hid_t lcpl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lcreate_soft(link_target, link_loc_id, link_name, lcpl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Ldelete(hid_t loc_id, char *name, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Ldelete(loc_id, name, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Ldelete_by_idx(hid_t loc_id, char *group_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Ldelete_by_idx(loc_id, group_name, idx_type, order, n, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lget_val(hid_t loc_id, char *name, void *bufout, size_t size, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lget_val(loc_id, name, bufout, size, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lget_val_by_idx(hid_t loc_id, char *group_name,  H5_index_t idx_type, H5_iter_order_t order, hsize_t n, void *bufout, size_t size, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lget_val_by_idx(loc_id, group_name, idx_type, order, n, bufout, size, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Lexists(hid_t loc_id, char *name, hid_t lapl_id) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Lexists(loc_id, name, lapl_id)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Lget_info(hid_t loc_id, char *name, H5L_info_t *linfo, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lget_info(loc_id, name, linfo, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lget_info_by_idx(hid_t loc_id, char *group_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, H5L_info_t *linfo, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lget_info_by_idx(loc_id, group_name, idx_type, order, n, linfo, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef ssize_t H5Lget_name_by_idx(hid_t loc_id, char *group_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, char *name, size_t size, hid_t lapl_id) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Lget_name_by_idx(loc_id, group_name, idx_type, order, n, name, size, lapl_id)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef herr_t H5Literate(hid_t grp_id, H5_index_t idx_type, H5_iter_order_t order, hsize_t *idx, H5L_iterate_t op, void *op_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Literate(grp_id, idx_type, order, idx, op, op_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Literate_by_name(hid_t loc_id, char *group_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t *idx, H5L_iterate_t op, void *op_data, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Literate_by_name(loc_id, group_name, idx_type, order, idx, op, op_data, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lvisit(hid_t grp_id, H5_index_t idx_type, H5_iter_order_t order, H5L_iterate_t op, void *op_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lvisit(grp_id, idx_type, order, op, op_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lvisit_by_name(hid_t loc_id, char *group_name, H5_index_t idx_type, H5_iter_order_t order, H5L_iterate_t op, void *op_data, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lvisit_by_name(loc_id, group_name, idx_type, order, op, op_data, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lunpack_elink_val(void *ext_linkval, size_t link_size, unsigned *flags, char **filename, char **obj_path) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lunpack_elink_val(ext_linkval, link_size, flags, filename, obj_path)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Lcreate_external(char *file_name, char *obj_name, hid_t link_loc_id, char *link_name, hid_t lcpl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Lcreate_external(file_name, obj_name, link_loc_id, link_name, lcpl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Oopen(hid_t loc_id, char *name, hid_t lapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Oopen(loc_id, name, lapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Oopen_by_addr(hid_t loc_id, haddr_t addr) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Oopen_by_addr(loc_id, addr)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Oopen_by_idx(hid_t loc_id, char *group_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, hid_t lapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Oopen_by_idx(loc_id, group_name, idx_type, order, n, lapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Oget_info(hid_t loc_id, H5O_info_t *oinfo) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oget_info(loc_id, oinfo)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Oget_info_by_name(hid_t loc_id, char *name, H5O_info_t *oinfo, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oget_info_by_name(loc_id, name, oinfo, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Oget_info_by_idx(hid_t loc_id, char *group_name,  H5_index_t idx_type, H5_iter_order_t order, hsize_t n, H5O_info_t *oinfo, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oget_info_by_idx(loc_id, group_name, idx_type, order, n, oinfo, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Olink(hid_t obj_id, hid_t new_loc_id, char *new_name, hid_t lcpl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Olink(obj_id, new_loc_id, new_name, lcpl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Ocopy(hid_t src_loc_id, char *src_name, hid_t dst_loc_id,  char *dst_name, hid_t ocpypl_id, hid_t lcpl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Ocopy(src_loc_id, src_name, dst_loc_id, dst_name, ocpypl_id, lcpl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Oincr_refcount(hid_t object_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oincr_refcount(object_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Odecr_refcount(hid_t object_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Odecr_refcount(object_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Oset_comment(hid_t obj_id, char *comment) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oset_comment(obj_id, comment)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Oset_comment_by_name(hid_t loc_id, char *name,  char *comment, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oset_comment_by_name(loc_id, name, comment, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef ssize_t H5Oget_comment(hid_t obj_id, char *comment, size_t bufsize) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Oget_comment(obj_id, comment, bufsize)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef ssize_t H5Oget_comment_by_name(hid_t loc_id, char *name, char *comment, size_t bufsize, hid_t lapl_id) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Oget_comment_by_name(loc_id, name, comment, bufsize, lapl_id)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef herr_t H5Ovisit(hid_t obj_id, H5_index_t idx_type, H5_iter_order_t order,  H5O_iterate_t op, void *op_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Ovisit(obj_id, idx_type, order, op, op_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Ovisit_by_name(hid_t loc_id, char *obj_name, H5_index_t idx_type, H5_iter_order_t order, H5O_iterate_t op, void *op_data, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Ovisit_by_name(loc_id, obj_name, idx_type, order, op, op_data, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Oclose(hid_t object_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Oclose(object_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Pcreate(hid_t plist_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Pcreate(plist_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Pcopy(hid_t plist_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Pcopy(plist_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef int H5Pget_class(hid_t plist_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Pget_class(plist_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Pclose(hid_t plist_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pclose(plist_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Pequal( hid_t id1, hid_t id2 ) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Pequal(id1, id2)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Pclose_class(hid_t id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pclose_class(id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_version(hid_t plist, unsigned int *super_, unsigned int* freelist,  unsigned int *stab, unsigned int *shhdr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_version(plist, super_, freelist, stab, shhdr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_userblock(hid_t plist, hsize_t size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_userblock(plist, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_userblock(hid_t plist, hsize_t * size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_userblock(plist, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_sizes(hid_t plist, size_t sizeof_addr, size_t sizeof_size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_sizes(plist, sizeof_addr, sizeof_size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_sizes(hid_t plist, size_t *sizeof_addr, size_t *sizeof_size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_sizes(plist, sizeof_addr, sizeof_size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_sym_k(hid_t plist, unsigned int ik, unsigned int lk) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_sym_k(plist, ik, lk)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_sym_k(hid_t plist, unsigned int *ik, unsigned int *lk) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_sym_k(plist, ik, lk)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_istore_k(hid_t plist, unsigned int ik) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_istore_k(plist, ik)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_istore_k(hid_t plist, unsigned int *ik) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_istore_k(plist, ik)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fclose_degree(hid_t fapl_id, H5F_close_degree_t fc_degree) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fclose_degree(fapl_id, fc_degree)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_fclose_degree(hid_t fapl_id, H5F_close_degree_t *fc_degree) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_fclose_degree(fapl_id, fc_degree)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fapl_core( hid_t fapl_id, size_t increment, hbool_t backing_store) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fapl_core(fapl_id, increment, backing_store)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_fapl_core( hid_t fapl_id, size_t *increment, hbool_t *backing_store) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_fapl_core(fapl_id, increment, backing_store)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fapl_family( hid_t fapl_id,  hsize_t memb_size, hid_t memb_fapl_id ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fapl_family(fapl_id, memb_size, memb_fapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_fapl_family( hid_t fapl_id, hsize_t *memb_size, hid_t *memb_fapl_id ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_fapl_family(fapl_id, memb_size, memb_fapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_family_offset( hid_t fapl_id, hsize_t offset) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_family_offset(fapl_id, offset)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_family_offset( hid_t fapl_id, hsize_t *offset) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_family_offset(fapl_id, offset)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fapl_log(hid_t fapl_id, char *logfile, unsigned int flags, size_t buf_size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fapl_log(fapl_id, logfile, flags, buf_size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fapl_multi(hid_t fapl_id, H5FD_mem_t *memb_map, hid_t *memb_fapl, char **memb_name, haddr_t *memb_addr, hbool_t relax) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fapl_multi(fapl_id, memb_map, memb_fapl, memb_name, memb_addr, relax)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_cache(hid_t plist_id, int mdc_nelmts, int rdcc_nelmts,  size_t rdcc_nbytes, double rdcc_w0) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_cache(plist_id, mdc_nelmts, rdcc_nelmts, rdcc_nbytes, rdcc_w0)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_cache(hid_t plist_id, int *mdc_nelmts, int *rdcc_nelmts, size_t *rdcc_nbytes, double *rdcc_w0) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_cache(plist_id, mdc_nelmts, rdcc_nelmts, rdcc_nbytes, rdcc_w0)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fapl_sec2(hid_t fapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fapl_sec2(fapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fapl_stdio(hid_t fapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fapl_stdio(fapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Pget_driver(hid_t fapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Pget_driver(fapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Pget_mdc_config(hid_t plist_id, H5AC_cache_config_t *config_ptr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_mdc_config(plist_id, config_ptr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_mdc_config(hid_t plist_id, H5AC_cache_config_t *config_ptr) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_mdc_config(plist_id, config_ptr)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_layout(hid_t plist, int layout) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_layout(plist, layout)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5D_layout_t H5Pget_layout(hid_t plist) except *:
    cdef H5D_layout_t r
    with rlock:
        r = _hdf5.H5Pget_layout(plist)
        if r<0:
            if set_exception():
                return <H5D_layout_t>-1;
        return r

cdef herr_t H5Pset_chunk(hid_t plist, int ndims, hsize_t * dim) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_chunk(plist, ndims, dim)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5Pget_chunk(hid_t plist, int max_ndims, hsize_t * dims ) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Pget_chunk(plist, max_ndims, dims)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Pset_deflate( hid_t plist, int level) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_deflate(plist, level)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fill_value(hid_t plist_id, hid_t type_id, void *value ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fill_value(plist_id, type_id, value)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_fill_value(hid_t plist_id, hid_t type_id, void *value ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_fill_value(plist_id, type_id, value)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pfill_value_defined(hid_t plist_id, H5D_fill_value_t *status ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pfill_value_defined(plist_id, status)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fill_time(hid_t plist_id, H5D_fill_time_t fill_time ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fill_time(plist_id, fill_time)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_fill_time(hid_t plist_id, H5D_fill_time_t *fill_time ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_fill_time(plist_id, fill_time)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_alloc_time(hid_t plist_id, H5D_alloc_time_t alloc_time ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_alloc_time(plist_id, alloc_time)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_alloc_time(hid_t plist_id, H5D_alloc_time_t *alloc_time ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_alloc_time(plist_id, alloc_time)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_filter(hid_t plist, H5Z_filter_t filter, unsigned int flags, size_t cd_nelmts, unsigned int* cd_values ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_filter(plist, filter, flags, cd_nelmts, cd_values)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Pall_filters_avail(hid_t dcpl_id) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Pall_filters_avail(dcpl_id)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef int H5Pget_nfilters(hid_t plist) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Pget_nfilters(plist)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef H5Z_filter_t H5Pget_filter(hid_t plist, unsigned int filter_number,   unsigned int *flags, size_t *cd_nelmts,  unsigned int* cd_values, size_t namelen, char* name ) except *:
    cdef H5Z_filter_t r
    with rlock:
        r = _hdf5.H5Pget_filter(plist, filter_number, flags, cd_nelmts, cd_values, namelen, name)
        if r<0:
            if set_exception():
                return <H5Z_filter_t>-1;
        return r

cdef herr_t H5Pget_filter_by_id( hid_t plist_id, H5Z_filter_t filter,  unsigned int *flags, size_t *cd_nelmts,  unsigned int* cd_values, size_t namelen, char* name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_filter_by_id(plist_id, filter, flags, cd_nelmts, cd_values, namelen, name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pmodify_filter(hid_t plist, H5Z_filter_t filter, unsigned int flags, size_t cd_nelmts, unsigned int *cd_values) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pmodify_filter(plist, filter, flags, cd_nelmts, cd_values)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Premove_filter(hid_t plist, H5Z_filter_t filter ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Premove_filter(plist, filter)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_fletcher32(hid_t plist) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_fletcher32(plist)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_shuffle(hid_t plist_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_shuffle(plist_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_szip(hid_t plist, unsigned int options_mask, unsigned int pixels_per_block) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_szip(plist, options_mask, pixels_per_block)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_scaleoffset(hid_t plist, H5Z_SO_scale_type_t scale_type, int scale_factor) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_scaleoffset(plist, scale_type, scale_factor)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_edc_check(hid_t plist, H5Z_EDC_t check) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_edc_check(plist, check)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5Z_EDC_t H5Pget_edc_check(hid_t plist) except *:
    cdef H5Z_EDC_t r
    with rlock:
        r = _hdf5.H5Pget_edc_check(plist)
        if r<0:
            if set_exception():
                return <H5Z_EDC_t>-1;
        return r

cdef herr_t H5Pset_chunk_cache( hid_t dapl_id, size_t rdcc_nslots, size_t rdcc_nbytes, double rdcc_w0 ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_chunk_cache(dapl_id, rdcc_nslots, rdcc_nbytes, rdcc_w0)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_chunk_cache( hid_t dapl_id, size_t *rdcc_nslots, size_t *rdcc_nbytes, double *rdcc_w0 ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_chunk_cache(dapl_id, rdcc_nslots, rdcc_nbytes, rdcc_w0)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_sieve_buf_size(hid_t fapl_id, size_t size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_sieve_buf_size(fapl_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_sieve_buf_size(hid_t fapl_id, size_t *size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_sieve_buf_size(fapl_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_nlinks(hid_t plist_id, size_t nlinks) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_nlinks(plist_id, nlinks)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_nlinks(hid_t plist_id, size_t *nlinks) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_nlinks(plist_id, nlinks)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_elink_prefix(hid_t plist_id, char *prefix) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_elink_prefix(plist_id, prefix)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef ssize_t H5Pget_elink_prefix(hid_t plist_id, char *prefix, size_t size) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Pget_elink_prefix(plist_id, prefix, size)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef hid_t H5Pget_elink_fapl(hid_t lapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Pget_elink_fapl(lapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Pset_elink_fapl(hid_t lapl_id, hid_t fapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_elink_fapl(lapl_id, fapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_create_intermediate_group(hid_t plist_id, unsigned crt_intmd) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_create_intermediate_group(plist_id, crt_intmd)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_create_intermediate_group(hid_t plist_id, unsigned *crt_intmd) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_create_intermediate_group(plist_id, crt_intmd)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_copy_object(hid_t plist_id, unsigned crt_intmd) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_copy_object(plist_id, crt_intmd)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_copy_object(hid_t plist_id, unsigned *crt_intmd) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_copy_object(plist_id, crt_intmd)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_char_encoding(hid_t plist_id, H5T_cset_t encoding) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_char_encoding(plist_id, encoding)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_char_encoding(hid_t plist_id, H5T_cset_t *encoding) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_char_encoding(plist_id, encoding)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_obj_track_times( hid_t ocpl_id, hbool_t track_times ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_obj_track_times(ocpl_id, track_times)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_obj_track_times( hid_t ocpl_id, hbool_t *track_times ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_obj_track_times(ocpl_id, track_times)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_local_heap_size_hint(hid_t plist_id, size_t size_hint) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_local_heap_size_hint(plist_id, size_hint)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_local_heap_size_hint(hid_t plist_id, size_t *size_hint) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_local_heap_size_hint(plist_id, size_hint)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_link_phase_change(hid_t plist_id, unsigned max_compact, unsigned min_dense) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_link_phase_change(plist_id, max_compact, min_dense)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_link_phase_change(hid_t plist_id, unsigned *max_compact , unsigned *min_dense) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_link_phase_change(plist_id, max_compact, min_dense)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_est_link_info(hid_t plist_id, unsigned est_num_entries, unsigned est_name_len) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_est_link_info(plist_id, est_num_entries, est_name_len)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_est_link_info(hid_t plist_id, unsigned *est_num_entries , unsigned *est_name_len) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_est_link_info(plist_id, est_num_entries, est_name_len)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_link_creation_order(hid_t plist_id, unsigned crt_order_flags) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_link_creation_order(plist_id, crt_order_flags)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_link_creation_order(hid_t plist_id, unsigned *crt_order_flags) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_link_creation_order(plist_id, crt_order_flags)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pset_libver_bounds(hid_t fapl_id, H5F_libver_t libver_low, H5F_libver_t libver_high) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pset_libver_bounds(fapl_id, libver_low, libver_high)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Pget_libver_bounds(hid_t fapl_id, H5F_libver_t *libver_low, H5F_libver_t *libver_high) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Pget_libver_bounds(fapl_id, libver_low, libver_high)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Rcreate(void *ref, hid_t loc_id, char *name, H5R_type_t ref_type,  hid_t space_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Rcreate(ref, loc_id, name, ref_type, space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Rdereference(hid_t obj_id, H5R_type_t ref_type, void *ref) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Rdereference(obj_id, ref_type, ref)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Rget_region(hid_t dataset, H5R_type_t ref_type, void *ref) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Rget_region(dataset, ref_type, ref)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef H5G_obj_t H5Rget_obj_type(hid_t id, H5R_type_t ref_type, void *ref) except *:
    cdef H5G_obj_t r
    with rlock:
        r = _hdf5.H5Rget_obj_type(id, ref_type, ref)
        if r<0:
            if set_exception():
                return <H5G_obj_t>-1;
        return r

cdef ssize_t H5Rget_name(hid_t loc_id, H5R_type_t ref_type, void *ref, char *name, size_t size) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Rget_name(loc_id, ref_type, ref, name, size)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef hid_t H5Screate(H5S_class_t type) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Screate(type)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Scopy(hid_t space_id ) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Scopy(space_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Sclose(hid_t space_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sclose(space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Screate_simple(int rank, hsize_t *dims, hsize_t *maxdims) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Screate_simple(rank, dims, maxdims)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef htri_t H5Sis_simple(hid_t space_id) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Sis_simple(space_id)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Soffset_simple(hid_t space_id, hssize_t *offset ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Soffset_simple(space_id, offset)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5Sget_simple_extent_ndims(hid_t space_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Sget_simple_extent_ndims(space_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Sget_simple_extent_dims(hid_t space_id, hsize_t *dims, hsize_t *maxdims) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Sget_simple_extent_dims(space_id, dims, maxdims)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef hssize_t H5Sget_simple_extent_npoints(hid_t space_id) except *:
    cdef hssize_t r
    with rlock:
        r = _hdf5.H5Sget_simple_extent_npoints(space_id)
        if r<0:
            if set_exception():
                return <hssize_t>-1;
        return r

cdef H5S_class_t H5Sget_simple_extent_type(hid_t space_id) except *:
    cdef H5S_class_t r
    with rlock:
        r = _hdf5.H5Sget_simple_extent_type(space_id)
        if r<0:
            if set_exception():
                return <H5S_class_t>-1;
        return r

cdef herr_t H5Sextent_copy(hid_t dest_space_id, hid_t source_space_id ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sextent_copy(dest_space_id, source_space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sset_extent_simple(hid_t space_id, int rank, hsize_t *current_size, hsize_t *maximum_size ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sset_extent_simple(space_id, rank, current_size, maximum_size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sset_extent_none(hid_t space_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sset_extent_none(space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5S_sel_type H5Sget_select_type(hid_t space_id) except *:
    cdef H5S_sel_type r
    with rlock:
        r = _hdf5.H5Sget_select_type(space_id)
        if r<0:
            if set_exception():
                return <H5S_sel_type>-1;
        return r

cdef hssize_t H5Sget_select_npoints(hid_t space_id) except *:
    cdef hssize_t r
    with rlock:
        r = _hdf5.H5Sget_select_npoints(space_id)
        if r<0:
            if set_exception():
                return <hssize_t>-1;
        return r

cdef herr_t H5Sget_select_bounds(hid_t space_id, hsize_t *start, hsize_t *end) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sget_select_bounds(space_id, start, end)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sselect_all(hid_t space_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sselect_all(space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sselect_none(hid_t space_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sselect_none(space_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Sselect_valid(hid_t space_id) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Sselect_valid(space_id)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef hssize_t H5Sget_select_elem_npoints(hid_t space_id) except *:
    cdef hssize_t r
    with rlock:
        r = _hdf5.H5Sget_select_elem_npoints(space_id)
        if r<0:
            if set_exception():
                return <hssize_t>-1;
        return r

cdef herr_t H5Sget_select_elem_pointlist(hid_t space_id, hsize_t startpoint,  hsize_t numpoints, hsize_t *buf) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sget_select_elem_pointlist(space_id, startpoint, numpoints, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sselect_elements(hid_t space_id, H5S_seloper_t op,  size_t num_elements, hsize_t **coord) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sselect_elements(space_id, op, num_elements, coord)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hssize_t H5Sget_select_hyper_nblocks(hid_t space_id ) except *:
    cdef hssize_t r
    with rlock:
        r = _hdf5.H5Sget_select_hyper_nblocks(space_id)
        if r<0:
            if set_exception():
                return <hssize_t>-1;
        return r

cdef herr_t H5Sget_select_hyper_blocklist(hid_t space_id,  hsize_t startblock, hsize_t numblocks, hsize_t *buf ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sget_select_hyper_blocklist(space_id, startblock, numblocks, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sselect_hyperslab(hid_t space_id, H5S_seloper_t op,  hsize_t *start, hsize_t *_stride, hsize_t *count, hsize_t *_block) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sselect_hyperslab(space_id, op, start, _stride, count, _block)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Sencode(hid_t obj_id, void *buf, size_t *nalloc) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Sencode(obj_id, buf, nalloc)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Sdecode(void *buf) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Sdecode(buf)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Tcreate(H5T_class_t type, size_t size) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tcreate(type, size)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Topen(hid_t loc, char* name) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Topen(loc, name)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Tcommit(hid_t loc_id, char* name, hid_t type) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tcommit(loc_id, name, type)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Tcommitted(hid_t type) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Tcommitted(type)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef hid_t H5Tcopy(hid_t type_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tcopy(type_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef htri_t H5Tequal(hid_t type_id1, hid_t type_id2 ) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Tequal(type_id1, type_id2)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Tlock(hid_t type_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tlock(type_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_class_t H5Tget_class(hid_t type_id) except *:
    cdef H5T_class_t r
    with rlock:
        r = _hdf5.H5Tget_class(type_id)
        if r<0:
            if set_exception():
                return <H5T_class_t>-1;
        return r

cdef size_t H5Tget_size(hid_t type_id) except *:
    cdef size_t r
    with rlock:
        r = _hdf5.H5Tget_size(type_id)
        if r==0:
            if set_exception():
                return <size_t>0;
        return r

cdef hid_t H5Tget_super(hid_t type) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tget_super(type)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef htri_t H5Tdetect_class(hid_t type_id, H5T_class_t dtype_class) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Tdetect_class(type_id, dtype_class)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Tclose(hid_t type_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tclose(type_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Tget_native_type(hid_t type_id, H5T_direction_t direction) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tget_native_type(type_id, direction)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Tconvert(hid_t src_id, hid_t dst_id, size_t nelmts, void *buf, void *background, hid_t plist_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tconvert(src_id, dst_id, nelmts, buf, background, plist_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tset_size(hid_t type_id, size_t size) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_size(type_id, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_order_t H5Tget_order(hid_t type_id) except *:
    cdef H5T_order_t r
    with rlock:
        r = _hdf5.H5Tget_order(type_id)
        if r<0:
            if set_exception():
                return <H5T_order_t>-1;
        return r

cdef herr_t H5Tset_order(hid_t type_id, H5T_order_t order) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_order(type_id, order)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hsize_t H5Tget_precision(hid_t type_id) except *:
    cdef hsize_t r
    with rlock:
        r = _hdf5.H5Tget_precision(type_id)
        if r==0:
            if set_exception():
                return <hsize_t>0;
        return r

cdef herr_t H5Tset_precision(hid_t type_id, size_t prec) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_precision(type_id, prec)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5Tget_offset(hid_t type_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Tget_offset(type_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Tset_offset(hid_t type_id, size_t offset) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_offset(type_id, offset)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tget_pad(hid_t type_id, H5T_pad_t * lsb, H5T_pad_t * msb ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tget_pad(type_id, lsb, msb)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tset_pad(hid_t type_id, H5T_pad_t lsb, H5T_pad_t msb ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_pad(type_id, lsb, msb)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_sign_t H5Tget_sign(hid_t type_id) except *:
    cdef H5T_sign_t r
    with rlock:
        r = _hdf5.H5Tget_sign(type_id)
        if r<0:
            if set_exception():
                return <H5T_sign_t>-1;
        return r

cdef herr_t H5Tset_sign(hid_t type_id, H5T_sign_t sign) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_sign(type_id, sign)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tget_fields(hid_t type_id, size_t *spos, size_t *epos, size_t *esize, size_t *mpos, size_t *msize ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tget_fields(type_id, spos, epos, esize, mpos, msize)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tset_fields(hid_t type_id, size_t spos, size_t epos, size_t esize, size_t mpos, size_t msize ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_fields(type_id, spos, epos, esize, mpos, msize)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef size_t H5Tget_ebias(hid_t type_id) except *:
    cdef size_t r
    with rlock:
        r = _hdf5.H5Tget_ebias(type_id)
        if r==0:
            if set_exception():
                return <size_t>0;
        return r

cdef herr_t H5Tset_ebias(hid_t type_id, size_t ebias) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_ebias(type_id, ebias)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_norm_t H5Tget_norm(hid_t type_id) except *:
    cdef H5T_norm_t r
    with rlock:
        r = _hdf5.H5Tget_norm(type_id)
        if r<0:
            if set_exception():
                return <H5T_norm_t>-1;
        return r

cdef herr_t H5Tset_norm(hid_t type_id, H5T_norm_t norm) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_norm(type_id, norm)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_pad_t H5Tget_inpad(hid_t type_id) except *:
    cdef H5T_pad_t r
    with rlock:
        r = _hdf5.H5Tget_inpad(type_id)
        if r<0:
            if set_exception():
                return <H5T_pad_t>-1;
        return r

cdef herr_t H5Tset_inpad(hid_t type_id, H5T_pad_t inpad) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_inpad(type_id, inpad)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_cset_t H5Tget_cset(hid_t type_id) except *:
    cdef H5T_cset_t r
    with rlock:
        r = _hdf5.H5Tget_cset(type_id)
        if r<0:
            if set_exception():
                return <H5T_cset_t>-1;
        return r

cdef herr_t H5Tset_cset(hid_t type_id, H5T_cset_t cset) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_cset(type_id, cset)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_str_t H5Tget_strpad(hid_t type_id) except *:
    cdef H5T_str_t r
    with rlock:
        r = _hdf5.H5Tget_strpad(type_id)
        if r<0:
            if set_exception():
                return <H5T_str_t>-1;
        return r

cdef herr_t H5Tset_strpad(hid_t type_id, H5T_str_t strpad) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_strpad(type_id, strpad)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Tvlen_create(hid_t base_type_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tvlen_create(base_type_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef htri_t H5Tis_variable_str(hid_t dtype_id) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Tis_variable_str(dtype_id)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef int H5Tget_nmembers(hid_t type_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Tget_nmembers(type_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef H5T_class_t H5Tget_member_class(hid_t type_id, int member_no) except *:
    cdef H5T_class_t r
    with rlock:
        r = _hdf5.H5Tget_member_class(type_id, member_no)
        if r<0:
            if set_exception():
                return <H5T_class_t>-1;
        return r

cdef char* H5Tget_member_name(hid_t type_id, unsigned membno) except *:
    cdef char* r
    with rlock:
        r = _hdf5.H5Tget_member_name(type_id, membno)
        if r==NULL:
            if set_exception():
                return <char*>NULL;
        return r

cdef hid_t H5Tget_member_type(hid_t type_id, unsigned membno) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tget_member_type(type_id, membno)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef int H5Tget_member_offset(hid_t type_id, int membno) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Tget_member_offset(type_id, membno)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Tget_member_index(hid_t type_id, char* name) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Tget_member_index(type_id, name)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Tinsert(hid_t parent_id, char *name, size_t offset, hid_t member_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tinsert(parent_id, name, offset, member_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tpack(hid_t type_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tpack(type_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Tenum_create(hid_t base_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tenum_create(base_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Tenum_insert(hid_t type, char *name, void *value) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tenum_insert(type, name, value)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tenum_nameof( hid_t type, void *value, char *name, size_t size ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tenum_nameof(type, value, name, size)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tenum_valueof( hid_t type, char *name, void *value ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tenum_valueof(type, name, value)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tget_member_value(hid_t type,  unsigned int memb_no, void *value ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tget_member_value(type, memb_no, value)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Tarray_create(hid_t base_id, int ndims, hsize_t *dims, int *perm) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tarray_create(base_id, ndims, dims, perm)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef int H5Tget_array_ndims(hid_t type_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Tget_array_ndims(type_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef int H5Tget_array_dims(hid_t type_id, hsize_t *dims, int *perm) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Tget_array_dims(type_id, dims, perm)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5Tset_tag(hid_t type_id, char* tag) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tset_tag(type_id, tag)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef char* H5Tget_tag(hid_t type_id) except *:
    cdef char* r
    with rlock:
        r = _hdf5.H5Tget_tag(type_id)
        if r==NULL:
            if set_exception():
                return <char*>NULL;
        return r

cdef hid_t H5Tdecode(unsigned char *buf) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Tdecode(buf)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Tencode(hid_t obj_id, unsigned char *buf, size_t *nalloc) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tencode(obj_id, buf, nalloc)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tcommit2(hid_t loc_id, char *name, hid_t dtype_id, hid_t lcpl_id, hid_t tcpl_id, hid_t tapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tcommit2(loc_id, name, dtype_id, lcpl_id, tcpl_id, tapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef H5T_conv_t H5Tfind(hid_t src_id, hid_t dst_id, H5T_cdata_t **pcdata) except *:
    cdef H5T_conv_t r
    with rlock:
        r = _hdf5.H5Tfind(src_id, dst_id, pcdata)
        if r==NULL:
            if set_exception():
                return <H5T_conv_t>NULL;
        return r

cdef herr_t H5Tregister(H5T_pers_t pers, char *name, hid_t src_id, hid_t dst_id, H5T_conv_t func) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tregister(pers, name, src_id, dst_id, func)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Tunregister(H5T_pers_t pers, char *name, hid_t src_id, hid_t dst_id, H5T_conv_t func) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Tunregister(pers, name, src_id, dst_id, func)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Zfilter_avail(H5Z_filter_t id_) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Zfilter_avail(id_)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Zget_filter_info(H5Z_filter_t filter_, unsigned int *filter_config_flags) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Zget_filter_info(filter_, filter_config_flags)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Acreate(hid_t loc_id, char *name, hid_t type_id, hid_t space_id, hid_t create_plist) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Acreate(loc_id, name, type_id, space_id, create_plist)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Aopen_idx(hid_t loc_id, unsigned int idx) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Aopen_idx(loc_id, idx)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Aopen_name(hid_t loc_id, char *name) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Aopen_name(loc_id, name)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Aclose(hid_t attr_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aclose(attr_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Adelete(hid_t loc_id, char *name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Adelete(loc_id, name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aread(hid_t attr_id, hid_t mem_type_id, void *buf) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aread(attr_id, mem_type_id, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Awrite(hid_t attr_id, hid_t mem_type_id, void *buf ) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Awrite(attr_id, mem_type_id, buf)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5Aget_num_attrs(hid_t loc_id) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5Aget_num_attrs(loc_id)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef ssize_t H5Aget_name(hid_t attr_id, size_t buf_size, char *buf) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5Aget_name(attr_id, buf_size, buf)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef hid_t H5Aget_space(hid_t attr_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Aget_space(attr_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef hid_t H5Aget_type(hid_t attr_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Aget_type(attr_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Aiterate(hid_t loc_id, unsigned * idx, H5A_operator_t op, void* op_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aiterate(loc_id, idx, op, op_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Adelete_by_name(hid_t loc_id, char *obj_name, char *attr_name, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Adelete_by_name(loc_id, obj_name, attr_name, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Adelete_by_idx(hid_t loc_id, char *obj_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Adelete_by_idx(loc_id, obj_name, idx_type, order, n, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hid_t H5Acreate_by_name(hid_t loc_id, char *obj_name, char *attr_name, hid_t type_id, hid_t space_id, hid_t acpl_id, hid_t aapl_id, hid_t lapl_id) except *:
    cdef hid_t r
    with rlock:
        r = _hdf5.H5Acreate_by_name(loc_id, obj_name, attr_name, type_id, space_id, acpl_id, aapl_id, lapl_id)
        if r<0:
            if set_exception():
                return <hid_t>-1;
        return r

cdef herr_t H5Aopen(hid_t obj_id, char *attr_name, hid_t aapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aopen(obj_id, attr_name, aapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aopen_by_name( hid_t loc_id, char *obj_name, char *attr_name, hid_t aapl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aopen_by_name(loc_id, obj_name, attr_name, aapl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aopen_by_idx(hid_t loc_id, char *obj_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, hid_t aapl_id, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aopen_by_idx(loc_id, obj_name, idx_type, order, n, aapl_id, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5Aexists_by_name( hid_t loc_id, char *obj_name, char *attr_name, hid_t lapl_id) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Aexists_by_name(loc_id, obj_name, attr_name, lapl_id)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef htri_t H5Aexists(hid_t obj_id, char *attr_name) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5Aexists(obj_id, attr_name)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5Arename(hid_t loc_id, char *old_attr_name, char *new_attr_name) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Arename(loc_id, old_attr_name, new_attr_name)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Arename_by_name(hid_t loc_id, char *obj_name, char *old_attr_name, char *new_attr_name, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Arename_by_name(loc_id, obj_name, old_attr_name, new_attr_name, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aget_info( hid_t attr_id, H5A_info_t *ainfo) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aget_info(attr_id, ainfo)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aget_info_by_name(hid_t loc_id, char *obj_name, char *attr_name, H5A_info_t *ainfo, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aget_info_by_name(loc_id, obj_name, attr_name, ainfo, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aget_info_by_idx(hid_t loc_id, char *obj_name, H5_index_t idx_type, H5_iter_order_t order, hsize_t n, H5A_info_t *ainfo, hid_t lapl_id) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aget_info_by_idx(loc_id, obj_name, idx_type, order, n, ainfo, lapl_id)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5Aiterate2(hid_t obj_id, H5_index_t idx_type, H5_iter_order_t order, hsize_t *n, H5A_operator2_t op, void *op_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5Aiterate2(obj_id, idx_type, order, n, op, op_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef hsize_t H5Aget_storage_size(hid_t attr_id) except *:
    cdef hsize_t r
    with rlock:
        r = _hdf5.H5Aget_storage_size(attr_id)
        if r==0:
            if set_exception():
                return <hsize_t>0;
        return r

IF MPI:
    cdef herr_t H5Pset_fapl_mpio(hid_t fapl_id, MPI_Comm comm, MPI_Info info) except *:
        cdef herr_t r
        with rlock:
            r = _hdf5.H5Pset_fapl_mpio(fapl_id, comm, info)
            if r<0:
                if set_exception():
                    return <herr_t>-1;
            return r
    
IF MPI:
    cdef herr_t H5Pset_fapl_mpiposix(hid_t fapl_id, MPI_Comm comm, hbool_t hints) except *:
        cdef herr_t r
        with rlock:
            r = _hdf5.H5Pset_fapl_mpiposix(fapl_id, comm, hints)
            if r<0:
                if set_exception():
                    return <herr_t>-1;
            return r
    
IF HDF5_VERSION >= (1, 8, 9):
    IF MPI:
        cdef herr_t H5Fset_mpi_atomicity(hid_t file_id, hbool_t flag) except *:
            cdef herr_t r
            with rlock:
                r = _hdf5.H5Fset_mpi_atomicity(file_id, flag)
                if r<0:
                    if set_exception():
                        return <herr_t>-1;
                return r
        
IF HDF5_VERSION >= (1, 8, 9):
    IF MPI:
        cdef herr_t H5Fget_mpi_atomicity(hid_t file_id, hbool_t *flag) except *:
            cdef herr_t r
            with rlock:
                r = _hdf5.H5Fget_mpi_atomicity(file_id, flag)
                if r<0:
                    if set_exception():
                        return <herr_t>-1;
                return r
        
cdef herr_t H5DSattach_scale(hid_t did, hid_t dsid, unsigned int idx) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5DSattach_scale(did, dsid, idx)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5DSdetach_scale(hid_t did, hid_t dsid, unsigned int idx) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5DSdetach_scale(did, dsid, idx)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef herr_t H5DSset_scale(hid_t dsid, char *dimname) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5DSset_scale(dsid, dimname)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef int H5DSget_num_scales(hid_t did, unsigned int dim) except *:
    cdef int r
    with rlock:
        r = _hdf5.H5DSget_num_scales(did, dim)
        if r<0:
            if set_exception():
                return <int>-1;
        return r

cdef herr_t H5DSset_label(hid_t did, unsigned int idx, char *label) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5DSset_label(did, idx, label)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef ssize_t H5DSget_label(hid_t did, unsigned int idx, char *label, size_t size) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5DSget_label(did, idx, label, size)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef ssize_t H5DSget_scale_name(hid_t did, char *name, size_t size) except *:
    cdef ssize_t r
    with rlock:
        r = _hdf5.H5DSget_scale_name(did, name, size)
        if r<0:
            if set_exception():
                return <ssize_t>-1;
        return r

cdef htri_t H5DSis_scale(hid_t did) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5DSis_scale(did)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r

cdef herr_t H5DSiterate_scales(hid_t did, unsigned int dim, int *idx, H5DS_iterate_t visitor, void *visitor_data) except *:
    cdef herr_t r
    with rlock:
        r = _hdf5.H5DSiterate_scales(did, dim, idx, visitor, visitor_data)
        if r<0:
            if set_exception():
                return <herr_t>-1;
        return r

cdef htri_t H5DSis_attached(hid_t did, hid_t dsid, unsigned int idx) except *:
    cdef htri_t r
    with rlock:
        r = _hdf5.H5DSis_attached(did, dsid, idx)
        if r<0:
            if set_exception():
                return <htri_t>-1;
        return r




# This file is part of h5py, a Python interface to the HDF5 library.
#
# http://www.h5py.org
#
# Copyright 2008-2013 Andrew Collette and contributors
#
# License:  Standard 3-clause BSD; see "license.txt" for full license terms
#           and contributor agreement.

try:
    import unittest2 as ut
except ImportError:
    import unittest as ut

import numpy as np
import h5py
from h5py import h5t

class TestCompound(ut.TestCase):

    """
        Feature: Compound types can be created from Python dtypes
    """

    def test_ref(self):
        """ Reference types are correctly stored in compound types (issue 144)
        """
        ref = h5py.special_dtype(ref=h5py.Reference)
        dt = np.dtype([('a',ref),('b','<f4')])
        tid = h5t.py_create(dt,logical=True)
        t1, t2 = tid.get_member_type(0), tid.get_member_type(1)
        self.assertEqual(t1, h5t.STD_REF_OBJ)
        self.assertEqual(t2, h5t.IEEE_F32LE)
        self.assertEqual(tid.get_member_offset(0), 0)
        self.assertEqual(tid.get_member_offset(1), h5t.STD_REF_OBJ.get_size())

